@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2018 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2019 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -51,9 +51,11 @@ class JobState(object):
         self.exception = None
         self.traceback = None
         self.tries = 0
+        self.etag = None
+        self.error_ignored = False
 
     def load(self):
-        self.old_data, self.timestamp, self.tries = self.cache_storage.load(self.job, self.job.get_guid())
+        self.old_data, self.timestamp, self.tries, self.etag = self.cache_storage.load(self.job, self.job.get_guid())
         if self.tries is None:
             self.tries = 0
 
@@ -62,39 +64,49 @@ class JobState(object):
             # If no new data has been retrieved due to an exception, use the old job data
             self.new_data = self.old_data
 
-        self.cache_storage.save(self.job, self.job.get_guid(), self.new_data, time.time(), self.tries)
+        self.cache_storage.save(self.job, self.job.get_guid(), self.new_data, time.time(), self.tries, self.etag)
 
     def process(self):
         logger.info('Processing: %s', self.job)
         try:
-            self.load()
-            data = self.job.retrieve(self)
+            try:
+                self.load()
+                data = self.job.retrieve(self)
 
-            # Apply automatic filters first
-            data = FilterBase.auto_process(self, data)
+                # Apply automatic filters first
+                data = FilterBase.auto_process(self, data)
 
-            # Apply any specified filters
-            filter_list = self.job.filter
+                # Apply any specified filters
+                filter_list = self.job.filter
 
-            if filter_list is not None:
-                if isinstance(filter_list, list):
-                    for item in filter_list:
-                        key = next(iter(item))
-                        filter_kind, subfilter = key, item[key]
-                        data = FilterBase.process(filter_kind, subfilter, self, data)
-                elif isinstance(filter_list, str):
-                    for filter_kind in filter_list.split(','):
-                        if ':' in filter_kind:
-                            filter_kind, subfilter = filter_kind.split(':', 1)
-                        else:
-                            subfilter = None
-                        data = FilterBase.process(filter_kind, subfilter, self, data)
-            self.new_data = data
-            self.tries = 0
+                if filter_list:
+                    if isinstance(filter_list, list):
+                        for item in filter_list:
+                            key = next(iter(item))
+                            filter_kind, subfilter = key, item[key]
+                            data = FilterBase.process(filter_kind, subfilter, self, data)
+                    elif isinstance(filter_list, str):
+                        for filter_kind in filter_list.split(','):
+                            if ':' in filter_kind:
+                                filter_kind, subfilter = filter_kind.split(':', 1)
+                            else:
+                                subfilter = None
+                            data = FilterBase.process(filter_kind, subfilter, self, data)
+                self.new_data = data
 
+            except Exception as e:
+                # job has a chance to format and ignore its error
+                self.exception = e
+                self.traceback = self.job.format_error(e, traceback.format_exc())
+                self.error_ignored = self.job.ignore_error(e)
+                if not (self.error_ignored or isinstance(e, NotModifiedError)):
+                    self.tries += 1
+                    logger.debug('Increasing number of tries to %i for %s', self.tries, self.job)
         except Exception as e:
+            # job failed its chance to handle error
             self.exception = e
             self.traceback = traceback.format_exc()
+            self.error_ignored = False
             if not isinstance(e, NotModifiedError):
                 self.tries += 1
                 logger.debug('Increasing number of tries to %i for %s', self.tries, self.job)
@@ -111,7 +123,9 @@ class Report(object):
 
     def _result(self, verb, job_state):
         if job_state.exception is not None:
-            logger.debug('Got exception while processing %r: %s', job_state.job, job_state.exception)
+            # TODO: Once we require Python >= 3.5, we can just pass in job_state.exception as "exc_info" parameter
+            exc_info = (type(job_state.exception), job_state.exception, job_state.exception.__traceback__)
+            logger.debug('Got exception while processing %r', job_state.job, exc_info=exc_info)
 
         job_state.verb = verb
         self.job_states.append(job_state)
