@@ -1,6 +1,6 @@
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2018 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2019 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -28,12 +28,17 @@
 
 
 import difflib
+import tempfile
+import subprocess
+import re
+import shlex
 import email.utils
 import itertools
 import logging
 import sys
 import time
 import cgi
+import functools
 
 import requests
 
@@ -55,6 +60,11 @@ except ImportError:
 logger = logging.getLogger(__name__)
 
 
+# Regular expressions that match the added/removed markers of GNU wdiff output
+WDIFF_ADDED_RE = r'[{][+].*?[+][}]'
+WDIFF_REMOVED_RE = r'[\[][-].*?[-][]]'
+
+
 class ReporterBase(object, metaclass=TrackSubClasses):
     __subclasses__ = {}
 
@@ -98,6 +108,21 @@ class ReporterBase(object, metaclass=TrackSubClasses):
         raise NotImplementedError()
 
     def unified_diff(self, job_state):
+        if job_state.job.diff_tool is not None:
+            with tempfile.NamedTemporaryFile() as old_file, tempfile.NamedTemporaryFile() as new_file:
+                old_file.write(job_state.old_data.encode('utf-8'))
+                old_file.flush()
+                new_file.write(job_state.new_data.encode('utf-8'))
+                new_file.flush()
+                cmdline = shlex.split(job_state.job.diff_tool) + [old_file.name, new_file.name]
+                proc = subprocess.Popen(cmdline, stdout=subprocess.PIPE)
+                stdout, _ = proc.communicate()
+                # Diff tools return 0 for "nothing changed" or 1 for "files differ", anything else is an error
+                if proc.returncode in (0, 1):
+                    return stdout.decode('utf-8')
+                else:
+                    raise subprocess.CalledProcessError(result, cmdline)
+
         timestamp_old = email.utils.formatdate(job_state.timestamp, localtime=1)
         timestamp_new = email.utils.formatdate(time.time(), localtime=1)
         return ''.join(difflib.unified_diff([l + '\n' for l in job_state.old_data.splitlines()],
@@ -145,7 +170,7 @@ class HtmlReporter(ReporterBase):
         for job_state in self.report.get_filtered_job_states(self.job_states):
             job = job_state.job
 
-            if job.__kind__ == 'url':
+            if job.LOCATION_IS_URL:
                 title = '<a href="{location}">{pretty_name}</a>'
             elif job.pretty_name() != job.get_location():
                 title = '<span title="{location}">{pretty_name}</span>'
@@ -216,6 +241,15 @@ class TextReporter(ReporterBase):
         show_details = cfg['details']
         show_footer = cfg['footer']
 
+        if cfg['minimal']:
+            for job_state in self.report.get_filtered_job_states(self.job_states):
+                pretty_name = job_state.job.pretty_name()
+                location = job_state.job.get_location()
+                if pretty_name != location:
+                    location = '%s (%s)' % (pretty_name, location)
+                yield ': '.join((job_state.verb.upper(), location))
+            return
+
         summary = []
         details = []
         for job_state in self.report.get_filtered_job_states(self.job_states):
@@ -281,8 +315,12 @@ class StdoutReporter(TextReporter):
 
     __kind__ = 'stdout'
 
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._has_color = sys.stdout.isatty() and self.config.get('color', False)
+
     def _incolor(self, color_id, s):
-        if sys.stdout.isatty() and self.config.get('color', False):
+        if self._has_color:
             return '\033[9%dm%s\033[0m' % (color_id, s)
         return s
 
@@ -298,7 +336,15 @@ class StdoutReporter(TextReporter):
     def _blue(self, s):
         return self._incolor(4, s)
 
+    def _get_print(self):
+        if sys.platform == 'win32' and self._has_color:
+            from colorama import AnsiToWin32
+            return functools.partial(print, file=AnsiToWin32(sys.stdout).stream)
+        return print
+
     def submit(self):
+        print = self._get_print()
+
         cfg = self.report.config['report']['text']
         line_length = cfg['line_length']
 
@@ -306,6 +352,10 @@ class StdoutReporter(TextReporter):
         body = '\n'.join(super().submit())
 
         for line in body.splitlines():
+            # Basic colorization for wdiff-style differences
+            line = re.sub(WDIFF_ADDED_RE, lambda x: self._green(x.group(0)), line)
+            line = re.sub(WDIFF_REMOVED_RE, lambda x: self._red(x.group(0)), line)
+
             # FIXME: This isn't ideal, but works for now...
             if line in separators:
                 print(line)
@@ -351,8 +401,6 @@ class EMailReporter(TextReporter):
         else:
             logger.error('Invalid entry for method {method}'.format(method=self.config['method']))
 
-        # TODO set_password(options.email_smtp, options.email_from)
-
         if self.config['html']:
             body_html = '\n'.join(self.convert(HtmlReporter).submit())
 
@@ -402,7 +450,9 @@ class PushoverReport(WebServiceReporter):
         return app.get_user(self.config['user'])
 
     def web_service_submit(self, service, title, body):
-        msg = service.create_message(title=title, message=body, html=True, sound='spacealarm')
+        sound = self.config['sound']
+        device = self.config['device']
+        msg = service.create_message(title=title, message=body, html=True, sound=sound, device=device)
         msg.send()
 
 
@@ -419,17 +469,24 @@ class PushbulletReport(WebServiceReporter):
 
 
 class MailGunReporter(TextReporter):
-    """Custom email reporter that use mailgun service"""
+    """Custom email reporter that uses Mailgun"""
 
     __kind__ = 'mailgun'
 
     def submit(self):
+        region = self.config.get('region', '')
         domain = self.config['domain']
         api_key = self.config['api_key']
         from_name = self.config['from_name']
         from_mail = self.config['from_mail']
         to = self.config['to']
 
+        if region == 'us':
+            region = ''
+
+        if region != '':
+            region = ".{0}".format(region)
+
         filtered_job_states = list(self.report.get_filtered_job_states(self.job_states))
         subject_args = {
             'count': len(filtered_job_states),
@@ -441,12 +498,12 @@ class MailGunReporter(TextReporter):
         body_html = '\n'.join(self.convert(HtmlReporter).submit())
 
         if not body_text:
-            logger.debug('Not calling mailgun API (no changes)')
+            logger.debug('Not calling Mailgun API (no changes)')
             return
 
-        logger.debug("Sending mailgun request for domain:'{0}'".format(domain))
+        logger.debug("Sending Mailgun request for domain:'{0}'".format(domain))
         result = requests.post(
-            "https://api.mailgun.net/v3/{0}/messages".format(domain),
+            "https://api{0}.mailgun.net/v3/{1}/messages".format(region, domain),
             auth=("api", api_key),
             data={"from": "{0} <{1}>".format(from_name, from_mail),
                   "to": to,
@@ -457,7 +514,7 @@ class MailGunReporter(TextReporter):
         try:
             json_res = result.json()
 
-            if (result.status_code == 200):
+            if (result.status_code == requests.codes.ok):
                 logger.info("Mailgun response: id '{0}'. {1}".format(json_res['id'], json_res['message']))
             else:
                 logger.error("Mailgun error: {0}".format(json_res['message']))
@@ -478,7 +535,8 @@ class TelegramReporter(TextReporter):
     def submit(self):
 
         bot_token = self.config['bot_token']
-        chat_id = self.config['chat_id']
+        chat_ids = self.config['chat_id']
+        chat_ids = [chat_ids] if isinstance(chat_ids, str) else chat_ids
 
         text = '\n'.join(super().submit())
 
@@ -487,9 +545,11 @@ class TelegramReporter(TextReporter):
             return
 
         result = None
-
         for chunk in self.chunkstring(text, self.MAX_LENGTH):
-            result = self.submitToTelegram(bot_token, chat_id, chunk)
+            for chat_id in chat_ids:
+                res = self.submitToTelegram(bot_token, chat_id, chunk)
+                if res.status_code != requests.codes.ok or res is None:
+                    result = res
 
         return result
 
@@ -501,7 +561,7 @@ class TelegramReporter(TextReporter):
         try:
             json_res = result.json()
 
-            if (result.status_code == 200):
+            if (result.status_code == requests.codes.ok):
                 logger.info("Telegram response: ok '{0}'. {1}".format(json_res['ok'], json_res['result']))
             else:
                 logger.error("Telegram error: {0}".format(json_res['description']))
@@ -513,3 +573,44 @@ class TelegramReporter(TextReporter):
 
     def chunkstring(self, string, length):
         return (string[0 + i:length + i] for i in range(0, len(string), length))
+
+
+class SlackReporter(TextReporter):
+    """Custom Slack reporter"""
+    MAX_LENGTH = 4096
+
+    __kind__ = 'slack'
+
+    def submit(self):
+        webhook_url = self.config['webhook_url']
+        text = '\n'.join(super().submit())
+
+        if not text:
+            logger.debug('Not calling slack API (no changes)')
+            return
+
+        result = None
+        for chunk in self.chunkstring(text, self.MAX_LENGTH):
+            res = self.submit_to_slack(webhook_url, chunk)
+            if res.status_code != requests.codes.ok or res is None:
+                result = res
+
+        return result
+
+    def submit_to_slack(self, webhook_url, text):
+        logger.debug("Sending slack request with text:{0}".format(text))
+        post_data = {"text": text}
+        result = requests.post(webhook_url, json=post_data)
+        try:
+            if result.status_code == requests.codes.ok:
+                logger.info("Slack response: ok")
+            else:
+                logger.error("Slack error: {0}".format(result.text))
+        except ValueError:
+            logger.error(
+                "Failed to parse slack response. HTTP status code: {0}, content: {1}".format(result.status_code,
+                                                                                             result.content))
+        return result
+
+    def chunkstring(self, string, length):
+        return (string[0 + i:length + i] for i in range(0, len(string), length))
